/*
    This file is part of the oolaboola digital dj program

    Copyright (C) 1999  Eric S. Tiedemann

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*

DESCRIPTION

sexpr representation and i/o in C++

motivated somewhat by scp (symbolic control protocol)

exceptions: DomainError UnexpectedEOF SyntaxError

copies are done deeply


*/

extern "C"
{
#include <errno.h>
}

#include "sexpr.h"

istream&
operator >>(istream& i, sexpr& s)
{
  s = sexpr::read(i);
  return i;
}

ostream&
operator <<(ostream& o, sexpr& s)
{
  sexpr::write(s, o);
  return o;
}

map<string, sexpr *> sexpr::symtab;

ostream &sexpr::write(sexpr &s, ostream& o)
{
  switch (s.t)
    {
    case NIL: 
      o << "()";
      break;

    case EOFO:
      o << "#<EOF OBJECT>";
      break;

    case STRING:
      o << '\"';
      
      for(size_t i = 0; i < s.s.size(); i++)
	{
	  switch (s.s[i])
	    {
	    case '\"': o << "\\\""; break;
	    case '\\': o << "\\\\"; break;
	    default: o << s.s[i]; break;
	    }
	}
      
      o << '\"';

      break;

    case SYMBOL:
      o << s.s;
      break;

    case FIX:
      o << s.v.i;
      break;

    case FLOAT:
      o << s.v.f;
      break;

    case PAIR:
      o << '(';
      write(*s.v.p.car, o);
      write_list_tail(*s.v.p.cdr, o);
      break;
      
    default:
      o << "#<WEIRD OBJECT>";
      break;
    }

  return o;
}

sexpr
sexpr::string2symbol(string s)
{
  return sexpr(s, true);
}

sexpr
sexpr::string2number(string s)
{
  errno = 0;
  const char *p = s.c_str();
  char *endp = const_cast<char *>(p);

  // should check for ERANGE below

  // check for float first so we don't confuse the beginning of one
  // for a fix
  double d;
  if ((d = strtod(p, &endp)), endp != p)
    return sexpr(d);
  
  long l;
  if ((l = strtol(p, &endp, 10)), endp != p)
    return sexpr(l);
  
  return sexpr();
}

void
sexpr::write_list_tail(sexpr &l, ostream &o)
{
  if (l.nullp())
    o << ')';
  else if (l.pairp())
    {
      o << ' ';
      sexpr::write(l.car(), o);
      write_list_tail(l.cdr(), o);
    }
  else
    {
      o << " . ";
      sexpr::write(l, o);
      o << ')';
    }
}

bool
sexpr::terminatingp(int c)
{
  if (isspace(c))
    return true;

  switch (c)
    {
    case EOF:
    case '(':
    case ')':
    case '\"':
    case ';':
      return true;
    default:
      return false;
    }
}

sexpr
sexpr::read_refusing_EOF(istream& i)
{
  sexpr o = sexpr::read(i);

  if (o.eof_objectp())
    throw UnexpectedEOF();
  else
    return o;
}

char
sexpr::read_char_refusing_EOF(istream& i)
{
  int c = i.get();

  if (c == EOF)
    throw UnexpectedEOF();
  else
    return c;
}

sexpr
sexpr::read_list(istream &i)
{
  sexpr o = read1(i);

  if (o.eof_objectp())
    throw UnexpectedEOF();

  switch (o.t)
    {
    case CLOSE:
      return sexpr(NIL);
      break;

    case DOT:
      {
	sexpr o = read_refusing_EOF(i);
	sexpr e = read1(i);

	if (e.t == CLOSE)
	  return o;
	else
	  {
	    if (e.eof_objectp())
	      throw UnexpectedEOF();
	    else
	      throw SyntaxError(); // "Malformed dotted list syntax."
	  }
      }
      break;

    default:
      {
	sexpr b = sexpr::read_list(i);
	sexpr *x = new sexpr(o); // want auto_ptr()s here
	sexpr *y = new sexpr(b);
	return sexpr(x, y);
      }
    }
}

sexpr
sexpr::read_string(istream &i)
{
  string s;
  int c;

  for (c = read_char_refusing_EOF(i); 
       c != '\"';
       c = read_char_refusing_EOF(i))
    {
      if (c == '\\')
	s += read_char_refusing_EOF(i);
      else
	s += c;
    }

  return sexpr(s);
}

void
sexpr::consume_comment(istream &i)
{
  int c;

  while ((c = i.get()) != EOF && c != '\n')
    ;
}

sexpr
sexpr::read_atom(char c, istream &i)
{
  string s = read_atom_string(c, i);
  sexpr o;

  if (((o = string2number(s)), !o.nullp())
      || ((o = string2symbol(s)), !o.nullp()))
    return o;
  else
    throw SyntaxError();	// "Malformed atom"
}

string
sexpr::read_atom_string(char c, istream &i)
{
  string s;
  s += c;

  while (!terminatingp(c = i.get()))
    s += c;

  i.unget();

  return s;
}

sexpr
sexpr::read_dot(istream &i)
{
  if (terminatingp(i.peek()))
    return sexpr(sexpr::DOT);
  else
    return read_atom('.', i);
}

sexpr
sexpr::read1(istream& i)
{
  int c = i.get();

  switch (c)
    {
    case EOF: 
      return sexpr(EOFO);
      break;

    case '(':
      return read_list(i);
      break;

    case ')':
      return sexpr(CLOSE);
      break;

    case '.':
      return read_dot(i);
      break;

    case '\"':
      return read_string(i);
      break;

    case ';':
      consume_comment(i);
      return read1(i);
      break;

    default:

      if (isspace(c))
	return read1(i);
      else
	return read_atom(c, i);
    }
}

sexpr
sexpr::read(istream& i)
{
  sexpr o = read1(i);

  switch (o.t)
    {
    case DOT:
      throw SyntaxError();	// "Superfluous period found in input."
      break;

    case CLOSE:
      throw SyntaxError();	// "Superfluous close paren found in input."
      break;

    default:
      return o;
    }
}

bool
sexpr::eq(const sexpr &a, const sexpr &b)
{
  if (a.t != b.t)
    return false;

  switch (a.t)
    {
    case STRING:
      return a.s == b.s;
      break;
    case SYMBOL:
      return a.s == b.s;
      break;
    case PAIR:
      return &a == &b;
      break;
    case FIX:
      return a.v.i == b.v.i;
      break;
    case FLOAT:
      return a.v.f == b.v.f;
      break;
    default:
      return true;
      break;
    }
}

#ifdef TEST
int
main1(int argc, char **argv)
{
  sexpr s;

  while (cin >> s && !s.eof_objectp())
    cout << s;

  return 0;
}

int
main(int argc, char **argv)
{
  try
    {
      return main1(argc, argv);
    }
  catch (exception &e)
    {
      cerr << "Unhandled exception: " << e.what() << "\n";
      return 1;
    }
  catch (...)
    {
      cerr << "Unknown unhandled exception!\n";
      return 1;
    }
      
  return 0;
}
#endif
